"""
Processing of NMR data-sets

NPK - the python front end for processing NMR data-sets
=======================================================

NPK is mostly a library for NMR processing. However It comes with a minimum user Interface which is presented here.
It consists in a shell script wrapper which can be called from the terminal, and which is called NPK.
This shell script transfers the control to python via the Interface.py python code.

USAGE

    first form :
        NPK
            for an interactive session.

        NPK < program.py
            is fine

        In this mode, the complete NPK space name is loaded, and you can use directly all the command described in the documentation.

    second form :
        for running an NMR processing directly from the comand line.
        
        NPK ACTION input_datafile output_datafile { processing_parameterfile } { additional parameters }

        where
            ACTION is the processing to apply
            input_datafile and output_datafile are Gifa binary files
            processing_parameterfile contains all the needed parameters for the processing
                is optionnal, if omited, default values will be systematically used
                however, it must be present (even empty) if additional parameters are to be used.
        In this mode, all the details are handled by the parameter files.
        You will find the complete documentation for the parameters in the respective doc files.

ACTION
    possible actions are :
        FT1D        - perfoms 1D Fourier Analysis
                      see details in Process1D documentation
        MaxEnt1D    - perfoms 1D Maximum Entropy Analysis
                      see details in Process1D documentation
        FT2D        - perfoms 2D Fourier Analysis
                      see details in Process2D documentation
        MaxEnt2D    - perfoms 2D Maximum Entropy Analysis
                      see details in Process2D documentation
        PreDOSY2D   - prepares for DOSY analysis
                      see details in Process2D documentation
        DOSY2D      - performs a DOSY analysis
                      see details in ProcessDosy documentation
        CHECK       - no action, permits to chek syntax
        help        - this help file

NMR BINARY FILES
    NPK reads only Gifa files.
    These files are read and created by the various versions of the Gifa program
        Gifa v4, see http://abcis.cbs.cnrs.fr
    and Gifa v5, see http://www.NMRtec.com
    The programm NMRNoteBook, developped by NMRtec is also able to export and import Gifa files
    
    File extensions :
    Back in Gifa v4 time, no extension was used to specify the NMR Gifa binary files.
    With Gifa V5 a new convention has been set-up for the file extensions :
            time domain     freq domain
    1D          .gf1            .gs1
    2D          .gf2            .gs2
    3D          .gf3            .gs3
    These extensions are now enforced by NPK
    It should be noted that no provision is made to specify a file in an intermediate state (processed in F2 but not in F1)
    it is up to you name these file depending on what you want to do with them afterwards.

    If you want to import NMR binary files from you spectrometer, you can
    - either use one of the Gifa v4 importers found on the WEB site (the util sub directory)
    - write a python importer
        for instance there is a Bruker import available in XXX

PARAMETER FILES

    The NMR processing is controled by parameters described in the respective documentation.
    Each and every parameter has a default value which is not too silly, so usually, giving no parameter may give you some result.
    However, you will have to modify the processing results by changing these parameters.

    Parameters are given either through a parameter file, or by inline commands.
    parameter files have the .gtb extention
    parameter files contains parameters in the form of a property list :

        #some comment
        param1=value string
        param2=value string    ... etc ...

    Addtional parameters can be given to the processing, independently to the parameter file
    by directly giving the parameter on the calling line, in the form additionnal_param=value
    
    additional parameterfiles are also associated to the binary data files
        input_datafile.gtb
            should be present near the input_datafile
            and contains details on the data not found in the binary gifa file format.
        output_datafile.gtb
            is created by the processing and contains processing details

EXAMPLES
    NPK FT1D input.gf1 output.gs1 

        launch the processing of a 1D FID, called input.gf1, and create the file output.gs1
        if the file is present, it is used as the processing parameter file, otherwise default parameters are used

    NPK FT2D input.gf2 output.gs2 processing.gtb

        launch the processing of a 2D experiment, called input.gf2, and create the 2D spectrum output.gs2
        the file processing.gtb is used as the processing parameter file.
        Note that this file may be empty.

    NPK DOSY2D Post_F2.gs2 DOSY.gs2 processing.gtb audittrail=test1.html  me_details=1  me_size=512

        launch the processing of a DOSY, on a dataset called Post_F2.gs2, and create the 2D DOSY spectrum DOSY.gs2
        the file processing.gtb is used as the processing parameter file.
        additionnal parameters are used :
            audittrail=test1.html overwrite the default audittrail file name
            me_details=1  me_size=512   overwrite the size of the Inverse Laplace processing defined in the processing.gtb file

STARTING NPK  & POSSIBLE ERRORS
    NPK requires a certain number of environment files to be present.
    $JAVA and $NPK_HOME are defined in the NPK shell script.

    The following sequence of event takes place when you start NPK 
    - a java virtual machine (JVM) must be available (found in $JAVA)
    - the jython code is launched in the JVM (found in $NPK_HOME/Java)
    - additional jython libraries are loaded (found in $NPK_HOME/python/Lib)
    - NPK code is initialized
        - loading the binary libraries (found in $NPK_HOME/Java)
        - loading and compiling the NPK python library (found in $NPK_HOME/python/NPK)
    - command line is parsed and processing starts ( found in $NPK_HOME/python/NPK/Interface.py )
    
See also:
    some additional code is provided in the Console code which is loaded when entering the interactive mode :
    
    - status()     : displays the internal state of the Kernel
    - import Graph : opens a minimum graphic interface to the 1D buffer (resize to redraw)

"""
def FT1D(audit,p_in,f_in,f_out,inputfilename,outputfilename):
    """FT processing of a 2D FID
    """
#    NPK FT1D /fichiers_RMN/1/data.gifa4 /fichiers_RMN/1/spectrum_py.gifa4 /fichiers_RMN/1/param_ft1D.gtb
    # Pre ft phase
    audittrail( audit, "phase", "FID preparation phase")
    pre_ft_1d( audit, inputfilename, "p1.gs1", p_in, f_in, f_out)
        
    # ft phase
    audittrail( audit, "phase", "Spectral analysis phase")
    ft_1d( audit, "memory", "p2.gs1", p_in, f_in, f_out)

    # Post ft phase
    audittrail( audit, "phase", "Post processing phase")
    post_ft_1d( audit, "memory", outputfilename, p_in, f_in, f_out)


def FT2D(audit,p_in,f_in,f_out,inputfilename,outputfilename):
    """FT processing of a 2D FID
    """
#    NPK FT2D /RMN_NMRtec/octaacetate/214/data.gifa4  /RMN_NMRtec/octaacetate/214/spectrum_py.gifa4  /RMN_NMRtec/octaacetate/214/param_ft2D.gtb
    # Preparation phase
    audittrail( audit, "phase", "FID analysis phase")
    pre_2d( audit, inputfilename, "memory", p_in, f_in, f_out)

    # Pre FT F2 phase
    audittrail( audit, "phase", "F2_axis - Pre FT phase")
    pre_ft_2df2( audit, "memory", "memory", p_in, f_in, f_out)

    # FT F2 phase
    audittrail( audit, "phase", "F2 axis - FT phase")
    ft_2df2( audit, "memory", "memory", p_in, f_in, f_out )

    # Post FT F2 phase
    audittrail( audit, "phase", "F2 axis - Post FT phase")
    post_ft_2df2( audit, "memory", "memory", p_in, f_in, f_out )

    # Pre FT F1 phase
    audittrail( audit, "phase", "F1_axis - Pre FT phase")
    pre_ft_2df1( audit, "memory", "memory", p_in, f_in, f_out )

    # FT F1 phase
    audittrail( audit, "phase", "F1 axis - FT phase")
    ft_2df1( audit, "memory", "memory", p_in, f_in, f_out )

    # Post FT F1 phase
    audittrail( audit, "phase", "F1 axis - Post FT phase")
    post_ft_2df1( audit, "memory", "memory", p_in, f_in, f_out )

    # Final phase
    audittrail( audit, "phase", "Final phase")
    post_2d( audit, "memory", outputfilename, p_in, f_in, f_out, "." )

def MaxEnt2D(audit,p_in,f_in,f_out,inputfilename,outputfilename):
    """MaxEnt processing of a 2D FID

    """
#    NPK MaxEnt2D  /RMN_NMRtec/octaacetate/214/data.gifa4 /RMN_NMRtec/octaacetate/214/maxent_py.gifa4 /RMN_NMRtec/octaacetate/214/param_me2D.gtb
    # Preparation phase
    audittrail( audit, "phase", "FID analysis phase")
    pre_2d( audit, inputfilename, "memory", p_in, f_in, f_out)

    # Pre FT F2 phase
    audittrail( audit, "phase", "F2_axis - Pre FT phase")
    pre_ft_2df2( audit, "memory", "memory", p_in, f_in, f_out)

    # Pre FT F1 phase
    audittrail( audit, "phase", "F1_axis - Pre FT phase")
    pre_ft_2df1( audit, "memory", "memory", p_in, f_in, f_out )


    # MaxEnt phase
    audittrail( audit, "phase", "2D MaxEnt phase")
    maxent_2d( audit, "memory", "memory", p_in, f_in, f_out )

    # Post FT F1 phase
    audittrail( audit, "phase", "2D Post MaxEnt phase")
    post_maxent_2d( audit, "memory", "memory", p_in, f_in, f_out )

    # Final phase
    audittrail( audit, "phase", "Final phase")
    post_2d( audit, "memory", outputfilename, p_in, f_in, f_out, "." )


def PreDosy2D(audit,p_in,f_in,f_out,inputfilename,outputfilename):
    """prepare a 2D for processing of a 2D DOSY experiment
    ends with a half processed file
    ** not yet debugged ! **
    """
    # Preparation phase
    audittrail( audit, "phase", "FID analysis phase")
    pre_2d( audit, inputfilename, "memory", p_in, f_in, f_out)

    # Pre FT F2 phase
    audittrail( audit, "phase", "F2_axis - Pre FT phase")
    pre_ft_2df2( audit, "memory", "memory", p_in, f_in, f_out)

    # FT F2 phase
    audittrail( audit, "phase", "F2 axis - FT phase")
    ft_2df2( audit, "memory", "memory", p_in, f_in, f_out )

    # Post FT F2 phase
    audittrail( audit, "phase", "F2 axis - Post FT phase")
    post_ft_2df2( audit, "memory", outputfilename, p_in, f_in, f_out )


def Dosy2D(audit,p_in,f_in,f_out,inputfilename,outputfilename):
    """processing of a 2D DOSY experiment
    
    starts with a half processed file
    """
    # Pre ILT phase
    audittrail( audit, "phase", "F1_axis - Pre ILT phase")
    pre_ilt_2df1( audit, inputfilename, "memory", p_in, f_in, f_out )

    # ILT phase
    audittrail( audit, "phase", "2D ILT phase")
    ilt_2df1( audit, "memory", "memory", p_in, f_in, f_out )

    # Post ILT F1 phase
    audittrail( audit, "phase", "2D Post ILT phase")
    post_ilt_2df1( audit, "memory", outputfilename, p_in, f_in, f_out )


import sys
import time
import re
from NPK import *

#### load parameters
sys.argv.pop(0)
try:
    action = sys.argv.pop(0)
    inputfilename = sys.argv.pop(0)
    outputfilename = sys.argv.pop(0)
    try:
        paramfilename = sys.argv.pop(0)
    except:
        paramfilename="param.gtb"
except:
    if (action.upper() != "HELP"):
        print "wrong number of arguments"
        print "\nArguments :"
        try:
            print ("action: "+str(action))
            print ("inputfilename: "+str(inputfilename))
            print ("outputfilename: "+str(outputfilename))
            print ("paramfilename: "+str(paramfilename))
        except:
            pass
        print "\ntry >NPK help\n"
        raise "wrong number of arguments\n"
    else:
         print __doc__
    sys.exit()

#start timer
t = time.clock()

# Open audit trail
auditfile =  outputfilename+".html"
# first check that the auditfile key is not present, and use if it's there
for i in range(1,len(sys.argv)):
        if sys.argv[i].startswith("auditfile="):
            l = re.split(r'(?<!\\)=', sys.argv[i],1)   #matches = but not \=
            auditfile = re.sub(r'\\=', '=',l[1])      # replaces \= by =

print "Audit file is -", auditfile
audit = auditinitial(auditfile,"NPK processing of "+str(action))
try:        # this try - finally is to insure that the audit file will be closed !
    audittrail( audit, "phase", "Loading parameters")

    # load parameters
    audittrail( audit, "text", "open processing parameters","filename",paramfilename)
    p_in=dict_load(paramfilename)

    # gather additionnal parameters on the calling line, in the form key=value
    while sys.argv:
        k = sys.argv.pop(0)
        (dkey,fval) = dict_parse(k)
        print "Additional property :", dkey, "=", fval
        try:            # check key existence
            val=p_in[dkey]
        except:
            pass
        else:
            print "WARNING, key -",dkey,"- defined twice"
            print "         previous value :",val
        p_in[dkey]=fval
        audittrail( audit, "text", "additional parameter", dkey,fval)

    audittrail( audit, "text", "open input file parameters")
    try:
        f_in=dict_load(inputfilename+".gtb")
    except:
        print "Warning, Input data file property empty"
        f_in={}

    f_out={}


    # mini parser
    if (action == "FT1D"):
        FT1D(audit,p_in,f_in,f_out,inputfilename,outputfilename)
    elif (action == "FT2D"):
        FT2D(audit,p_in,f_in,f_out,inputfilename,outputfilename)
    elif (action == "MaxEnt2D"):
        MaxEnt2D(audit,p_in,f_in,f_out,inputfilename,outputfilename)
    elif (action == "PreDOSY2D"):
        PreDosy2D(audit,p_in,f_in,f_out,inputfilename,outputfilename)
    elif (action == "DOSY2D"):
        Dosy2D(audit,p_in,f_in,f_out,inputfilename,outputfilename)
    elif (action == "CHECK"):
        print "CHECK command"
        print ("action: %s  input: %s  output: %s  param: %s"%(action,inputfilename,outputfilename,paramfilename))
        sys.exit()
    else:
        raise "Action -" + action + "- unknown"


finally:    # finally, dump parameters and exit
    audittrail( audit, "phase", "End of processing")
    dict_dump(f_out,outputfilename+".gtb")
    audittrail( audit, "text", "write output file parameters", "file name", outputfilename+".gtb")

    t = time.clock()-t
    audittrail( audit, "text", "total elapsed time", "time",str(t)+" sec")
    audittrail( audit, "close")
    audit.close()

print "total elapsed time", "time:" ,str(t)+" sec"
sys.exit()
