"""
This modules implements high level tools for handling Peaks.

The following high level Classes are defined:
    Peak1D : a Class to handle 1D peaks
    Peak2D : a Class to handle 1D peaks
    Fitter1D : Fits peaks on the 1D current spectrum
        The fitter uses fitMask, which are dumy peaks with only 1 5to be fitted) and 0 (not fitted) values

still lacking
    PickerxD : to create peaklists from the current spectrum
    Fitter2D

    This is still work in progress

    author : M-A Delsuc
    some code rewritten from Vincent Catherinot java code
    date : April 2008
"""
__author__ = "M-A Delsuc <ma.delsuc@cbs.cnrs.fr>"
__version__ = "0.2"


try:
    import nmrtec.nmrNotebook.kernel.Kore as Kore
except ImportError:
    print "The NPK dll seems to be lacking"
try:
    import nmrtec.nmrNotebook.kernel.Fitter as FitterCalc   # this is the external jni library
except ImportError:
    print "The Fitter dll seems to be lacking"


#########################################################################
# The axis class is used also by 1D, 2D and nD peaks
#########################################################################
class Axis:
    """
    A class to handle isotropic spectral axes, used by PeakxD.
    contains
    shape   The shape of the peak 1.0 for Lorentzian / 0.0 for Gaussian - intermediate values possible
    pos     the position of the peak along this axis, in Hz
    width   the width of the peak
    phase   the phase of the peak in degrees
    pos_err    the error on the previous quantities
    width_err
    phase_err
    left        The extension of the integration zone - in index
    right
    """

    #---------------------------------------------------------------------------
    def __init__(self, pos=0.0, width=1.0,  phase=0.0, shape=1.0, pos_err=0.0, width_err=0.0, phase_err=0.0, left=0.0, right=0.0):
        self.pos=pos  # position in Hz
        self.width=width # width in Hz
        self.phase=phase
        self.shape=shape
            # errors
        self.pos_err=pos_err
        self.width_err=width_err
        self.phase_err=phase_err
        self.left=left  # left and right contain extensions used for integration
        self.right=right
    #---------------------------------------------------------------------------
    def __str__(self):
        if self.shape == 1.0:
            sh = "Lorentzian"
        elif self.shape == 0.0:
            sh = "Gaussian  "   
        else:
            sh = "%d%% Loren. %d%% Gauss."%(int(100*self.shape),int(100*(1-self.shape)))
        #return "%s   pos: %f+/-%f   width: %f+/-%f   phase: %f+/-%f   left: %d  right: %d" % (sh, self.pos,self.pos_err,self.width,self.width_err,self.phase,self.phase_err,self.left, self.right) Modif RH 18-11-2008
        return "%s   pos: %f+:-%f   width: %f+:-%f   phase: %f+:-%f   left: %d  right: %d" % (sh, self.pos,self.pos_err,self.width,self.width_err,self.phase,self.phase_err,self.left, self.right)
    def table(self,sep=","):
        """returns a string with the peak values separated with 'sep'
        ideal for outputing to "excel" files"""
        return "%f%s%f%s%f%s%f%s%f%s%f%s%f%s%d%s%d" % (self.shape,sep,self.pos,sep,self.pos_err,sep,self.width,sep,self.width_err,sep,self.phase,sep,self.phase_err,sep,self.left,sep,self.right)
    def table_hdr(self,sep=","):
        """returns a string suitable to be used as a header for table()"""
        return "shape%spos%spos_err%swidth%swidth_err%sphase%sphase_err%sleft%sright" % (sep,sep,sep,sep,sep,sep,sep,sep)
        

#########################################################################
class PkCommon:
    """
    This class defines the parameters common to all PeakxD :
    label   The name of the peak
    amp     The amplitude (the height of the largest point)
    area    The area/volume of the peak
    amp_err The error of the previous values
    area_err
    """
    #---------------------------------------------------------------------------
    def __init__(self, label="Peak", amp=0.0, area=0.0, amp_err=0.0, area_err=0.0):
        self.label=label
        self.amp = amp   # Amplitude in AU
        self.area = area   # volume in AU
            # errors
        self.amp_err=amp_err
        self.area_err=area_err

#########################################################################
class Peak1D(PkCommon):
    """
    A class to handle 1D peaks ! 
    presents a clean and natural interface to peak handling
    communicate with the peaks in the kernel with getFromKernel and copyToKernel

    peak lists are naturally handled as list of Peak1D (!!!)

    """
#---------------------------------------------------------------------------
    def __init__(self, label="Peak", amp=0.0, pos=0.0, width=1.0,  phase=0.0, shape=1.0,  area=0.0, amp_err=0.0, pos_err=0.0, width_err=0.0, phase_err=0.0,  area_err=0.0, left=0.0, right=0.0):
        self.label=label
        self.amp = amp   # Amplitude in AU
        self.area = area   # volume in AU
            # errors
        self.amp_err=amp_err
        self.area_err=area_err
        self.F1=Axis(pos, width,  phase, shape, pos_err, width_err, phase_err, left, right)
    def __str__(self):
        return "<%s>  amp: %f+:-%f   area: %f+:-%f   %s"%(self.label,self.amp,self.amp_err,self.area,self.area_err, str(self.F1))
    def table(self,sep=","):
        """returns a string with the peak values separated with 'sep'
        ideal for outputing to "excel" files"""
        return '"%s"%s%f%s%f%s%f%s%f%s%s'%(self.label,sep,self.amp,sep,self.amp_err,sep,self.area,sep,self.area_err,sep, self.F1.table(sep))
    def table_hdr(self,sep=","):
        """returns a string suitable to be used as a header for table()"""
        return "Label%sAmplitude%sAmplitude_err%sArea%sArea_err%s%s"%(sep,sep,sep,sep,sep,self.F1.table_hdr(sep))
#---------------------------------------------------------------------------
    def getFromKernel(self,i):
        """
        copy the value for peak i in kernel into this peak
        values undefined by the kernel will be untouched
        """
        raise "yet to be done"
        self.label = Kore.geta_pk1d_id(i)
        self.amp = Kore.geta_pk1d_a(i)
        self.amp_err = Kore.geta_pk1d_a_err(i)
        self.F1.pos = Kore.geta_pk1d_f(i)
        self.F1.pos_err = Kore.geta_pk1d_f_err(i)
        self.F1.width = Kore.geta_pk1d_w(i)
        self.F1.width_err = Kore.geta_pk1d_w_err(i)
        self.F1.phase = Kore.geta_pk1d_p(i)
        self.F1.left = Kore.geta_pk1d_left(i)
        self.F1.right = Kore.geta_pk1d_right(i)
        t = Kore.geta_pk1d_t(i)
        if t == 'L':
            self.F1.shape = 1.0
        elif t == 'G':
            self.F1.shape = 0.0
#---------------------------------------------------------------------------
    def copyToKernel(self,i):
        """copy the value for this peak into the kernel"""
        pass
        raise "yet to be done"
#########################################################################
class Peak2D(PkCommon):
    """
    A class to handle 2D peaks ! 
    presents a clean and natural interface to peak handling
    communicate with the peaks in the kernel with getFromKernel and copyToKernel

    peak lists are naturally handled as list of Peak2D (!!!)

    """
#---------------------------------------------------------------------------
    def __init__(self, label="Peak", amp=0.0, F1=(0.0, 1.0, 0.0, 1.0, 0, 0), F2=(0.0, 1.0, 0.0, 1.0, 0, 0), area=0.0, amp_err=0.0, F1_err=(0.0, 0.0, 0.0),  F2_err=(0.0, 0.0, 0.0),  area_err=0.0 ):
        self.label=label
        self.amp = amp   # Amplitude in AU
        self.area = area   # volume in AU
        # axes
        (pos, width,  phase, shape, left, right)  = F1
        (pos_err, width_err,  phase_err)  = F1_err
        self.F1=Axis(pos, width,  phase, shape, pos_err, width_err, phase_err, left, right)
        (pos, width,  phase, shape, left, right)  = F2
        (pos_err, width_err,  phase_err)  = F2_err
        self.F2=Axis(pos, width,  phase, shape, pos_err, width_err, phase_err, left, right)
        # errors
        self.amp_err=amp_err
        self.area_err=area_err
    def __str__(self):
        return "<%s>  amp: %f+:-%f   area: %f+:-%f   F1: %s   F2: %s"%(self.label,self.amp,self.amp_err,self.area,self.area_err, str(self.F1), str(self.F2))
#---------------------------------------------------------------------------
    def getFromKernel(self,i):
        """
        copy the value for peak i in kernel into this peak
        values undefined by the kernel will be untouched
        """
        raise "yet to be done"
#---------------------------------------------------------------------------
    def copyToKernel(self,i):
        """copy the value for this peak into the kernel"""
        raise "yet to be done"

#########################################################################
class Fitter1D:
    """
    This class implements high level tools for fitting 1D Peaks.
    
    start, end determines the spectral window to fit in index
    pklist is the list of Peak to actually fit
    fitmasklist is a list of peaks, telling what to fit and what not to fit, if it is not given,
    a default mask is created and filled so that Amp and Position only are fitted
    possible values are : 
        FitA (amp)  FitP (Pos)   FitW (width)    FitP (phase)   FitS (Lorentz-Gauss shape)
        and combinations :  FitAP FitAPW FitAPWP FitAPWS FitAPWPS


    Results are returned in Fitter.FittedPeakList

    return of info and error estimate is still to be done.

    Fitter should be called with the spectrum to fit into the 1D buffer


    A typical call is :
        # initialize list of peak to fit
        pl=[]
        pl.append(Peak.Peak1D("Line 2", 6000000, 3507, 2.0))
        pl.append(Peak.Peak1D("Line 3", 6000000, 3499, 2.0))

        # initialize fitter
        # first two arg are spectral window to fit (in index)
        # then peak list
        # then default mask defining which parameters to fit
        fitter=Linefit.Fitter(1320, 1350, pl,Peaker1D.Fitter.FitAPW)

        # start fitter
        fitter.doFit()

        # get fitted values
        Peaker1D.report(fitter.FittedPeakList)

        F = Fitter()

        author : M-A Delsuc
        some code rewritten from Vincent Catherinot java code
        date : April 2008

    """
    # a few usefull definitions for fit masks
    FitA=(1,0,0,0,0)
    FitP=(0,1,0,0,0)
    FitW=(0,0,1,0,0)
    FitP=(0,0,0,1,0)
    FitS=(0,0,0,0,1)
    FitAP=(1,1,0,0,0)
    FitAPW=(1,1,1,0,0)
    FitAPWP=(1,1,1,1,0)
    FitAPWS=(1,1,1,0,1)
    FitAPWPS=(1,1,1,1,1)
#---------------------------------------------------------------------------
    def __init__(self, start=0, end=0, pklist=[], fitmask=FitAP):
        self.setPeakList(pklist,fitmask)
        self.win_start=start
        self.win_end=end
        self.result=[]  # will hold the return from a fit
        self.FittedPeaks=[]     # will hold the fitted peak list
        # retrieve some infos about the spectrum
        self.loadData()
        self.FittedPeakList=[]
#---------------------------------------------------------------------------
    def setPeakList(self, pklist, fitmask=FitAP):
        """
        set the list of peaks to fit,
        pklist is the list of Peak to actually fit
        fitmask tells what to fit and what not to fit, if it is not given, Amp and Position only are fitted
        """
        self.PeakList = pklist
        self.setFitMaskList(fitmask)
#---------------------------------------------------------------------------
    def setFitMaskList(self,fitmask):
        """
        set the list of the fitmask to be used for fitting, using the same mask for all peaks
        """
        fm=[]
        (amp,pos,width,phase,shape)=fitmask
        for pk in self.PeakList:
                fm.append(Peak1D("mask", amp,pos,width,phase,shape)) # only amp and pos
        self.fitMaskList=fm
#---------------------------------------------------------------------------
    def setFitMaskPeak(self,peak_index,fitmask=FitAP):
        """
        Set the mask of the peak_index^th peak in self.PeakList
        """
        if not self.fitMaskList:
            self.setFitMaskList(FitAP)
        self.fitMaskList[peak_index]=mask
#---------------------------------------------------------------------------
    def loadData(self):
        """
        Load current Kernel 1D buffer for fitting.
        This is done at Fitter initialisation, but can be redone any time.
        """
        # retrieve some infos about the spectrum
        self.nbf1 = Kore.get_si1_1d()
        self.sw = Kore.get_specw_1d()
        self.off = Kore.get_offset_1d()
        # strange offset of 1 half point
        self.decal = -0.5*self.sw/self.nbf1
#---------------------------------------------------------------------------
    def prepData(self):
        """
        Prepare the data in the 1D buffer by isolating and applying Hilbert transform
        """
        Kore.put("data")    # keep data
        Kore.one()      # build window
        si = Kore.get_si1_1d()
        Kore.chsize( si - 2*self.win_start )
        Kore.chsize(si)
        Kore.reverse()
        Kore.chsize(2*self.win_end)
        Kore.chsize(si)
        # apply the window
        Kore.multdata()
        # And apply hilbert transform
        Kore.real()
        Kore.iftbis()
        Kore.chsize(si)
        Kore.ft()
#---------------------------------------------------------------------------
    def prepPeaks(self,fit):
        """
        Prepare the peaks into memory
        PeakList is a list of Peaks
        returns the number of peak copied
        """
        i = 0
        for (peak,mask) in zip(self.PeakList,self.fitMaskList):
#            print "%d  %s\n   %s\n"%(i, peak.report(), mask.report())
            FitterCalc.fitter_addRay( fit, FitterCalc.RAY_MIXTE, 1.0, 0, 0, 0, 0)
            FitterCalc.fitter_setRayAmp( fit, i, peak.amp )   # set value
            FitterCalc.fitter_setRayPhase( fit, i, peak.F1.phase)
            FitterCalc.fitter_setRayPos( fit, i, peak.F1.pos - self.decal)
            FitterCalc.fitter_setRayWidth( fit, i, peak.F1.width)
            FitterCalc.fitter_setRayMix( fit, i, peak.F1.shape)
            if mask.amp != 0:    # set mask
#                    print "fitting amp for peak ",i
                FitterCalc.fitter_releaseParm( fit, i, FitterCalc.PARM_AMPR)
            else:
                FitterCalc.fitter_fixParm( fit, i, FitterCalc.PARM_AMPR)
            if mask.F1.phase != 0:
#                    print "fitting phase for peak ",i
                FitterCalc.fitter_releaseParm( fit, i, FitterCalc.PARM_AMPI)
            else:
                FitterCalc.fitter_fixParm( fit, i, FitterCalc.PARM_AMPI)
            if mask.F1.pos != 0:
#                    print "fitting pos for peak ",i
                FitterCalc.fitter_releaseParm( fit, i, FitterCalc.PARM_POS)
            else:
                FitterCalc.fitter_fixParm( fit, i, FitterCalc.PARM_POS)
            if mask.F1.width != 0:
#                    print "fitting width for peak ",i
                FitterCalc.fitter_releaseParm( fit, i, FitterCalc.PARM_WIDTH)
            else:
                FitterCalc.fitter_fixParm( fit, i, FitterCalc.PARM_WIDTH)
            if mask.F1.shape != 0:
#                    print "fitting shape for peak ",i
                FitterCalc.fitter_releaseParm( fit, i, FitterCalc.PARM_MIX)
            else:
              FitterCalc.fitter_fixParm( fit, i, FitterCalc.PARM_MIX)
            i=i+1
        return (i-1)
#---------------------------------------------------------------------------
    def FitSize(self):
        """
        returns the size of the fitting problem as a pair (npeak, npar)
        where 
            npeak is the number of peak to fit
            pvar is the number of parameters to fit
        npar can take any value from 0 (nothing to fit) to 6*npeak (fit everything)
        """
        npeak=len(self.PeakList)
        npar=0
        for m in self.fitMaskList:
            if m.amp:
                    npar=npar+1
            if m.F1.phase:
                    npar=npar+1
            if m.F1.pos:
                    npar=npar+1
            if m.F1.width:
                    npar=npar+1
            if m.F1.shape:
                    npar=npar+1
        return (npeak,npar)
#---------------------------------------------------------------------------
    def getPeaks(self,fit):
        """
        Get fitted peaks out of the fitter, and return as a list
        """
        pl=[]
        for i in range(len(self.PeakList)):
            pk=Peak1D()
            pk.amp = FitterCalc.fitter_getRayAmp( fit, i)
            pk.F1.phase = FitterCalc.fitter_getRayPhase( fit, i)
            pk.F1.shape = FitterCalc.fitter_getRayMix( fit, i)
            pk.F1.pos = FitterCalc.fitter_getRayPos( fit, i) + self.decal
            pk.F1.width = FitterCalc.fitter_getRayWidth( fit, i)
            pl.append(pk)
        return pl
#---------------------------------------------------------------------------
    def doFit(self):
        """
        realizes the fit, once all parameters are set
        """
        if ( self.win_start < 1 ):
            raise "win_start < 1"
        if ( self.win_end > Kore.get_si1_1d() ):
            raise "win_end > si1_1d"
        #  check if there are enough data to fit the variables
        (npeak,npar) = self.FitSize()
        if ( npar > self.win_end-self.win_start):
            raise("Fitter error, the window is not large enough to fit %d lines with %d parameters."%(npeak,npar))

        # prepare data in kernel buffer - 1D remains untouched in data buffer
        self.prepData()

        # allocate the object
        fit = FitterCalc.fitter_newFitter()

        ## CHANGED from java code
        # load spectrum
        szfit = Kore.get_si1_1d()
        FitterCalc.fitter_set1Dszfid( fit, szfit/2)
        FitterCalc.fitter_loadFID(fit)  # ?

        # set-up the context
        FitterCalc.fitter_set1Dspecw(fit, self.sw)
        FitterCalc.fitter_set1Doffset(fit, self.off)
        FitterCalc.fitter_set1Dt0(fit, 0.0)                # ?

        # no linop
        FitterCalc.fitter_clearLinop(fit)               # ?
        FitterCalc.fitter_addLinopWindow(fit, self.win_start, self.win_end-2)

        # transfert peaks to optimize, along with masks
        self.prepPeaks(fit)

        # start the fit
        self.result = FitterCalc.fitter_startminpack(fit)

        # retrieve the value of optimized parameters
        self.FittedPeakList=self.getPeaks(fit)

        # free memory
        FitterCalc.fitter_delFitter(fit)
        # restore 1D Kore buffer
        Kore.get("data")
        return self.result
    
#########################################################################
# Utilities
#########################################################################
def addFitMask(m1, m2):
    """
    add two fit masks and return the sum
    thus, FitAP = addFitMask(FitA,FitP)
    """
    mm=[]
    for (i,j) in zip(m1,m2):
        mm.append(min(i+j,1))    # min() insures 1+1 => 1
    return tuple(mm)

def report(PeakList):
    """
    Make a simple report of a peaklist
    """
    for i in range(len(PeakList)):
        print "Peak n.%d : %s" % (i, str(PeakList[i]))

def table(PeakList,filename="PeakTable.txt",sep=","):
    """
    Creates a file, name filename, containing a table of a peaklist
    Which will be readable in 'excel'-like programs
    using sep as a separator
    """
    import time
    F=open(filename,"w")
    F.write("# Peak-List  file, created :"+time.strftime("%a, %d %b %Y %H:%M:%S %Z", time.localtime())+"\n")
    F.write("id%s%s\n" % (sep, PeakList[0].table_hdr(sep)))
    for i in range(len(PeakList)):
        F.write("%d%s%s\n"%(i,sep,PeakList[i].table(sep)))
    F.close()




if (__name__ == "__main__"):
    print "################ testing Peaks ############"
    P=Peak1D()
    print (P)
    P2=Peak2D()
    print (P2)
    print "################ testing Fitter1D ############"
    Kore.readc("C:/Documents and Settings/User/Bureau/MAD_14_11_2008/fit_decal/Lasalocid.gs1")
    #Kore.readc("/Users/mad/Documents/Developpement/NPK/examples/linefit/Lasalocid.gs1")
    pl=[]
    pl.append(Peak1D("Ray 4", 6000000, 3507, 2.0))
    pl.append(Peak1D("Ray 3", 6000000, 3499, 2.0))
    print "### initial table & mask "
    report(pl)
    fitter=Fitter1D(1320, 1350, pl,Fitter1D.FitAPW)
    report(fitter.fitMaskList)
    print "problem Size:", fitter.FitSize()
    fitter.doFit()
    print "### final (fitted) table "
    report(fitter.FittedPeakList)
